import torch


class PDELU(torch.nn.Module):
    def __init__(self, theta: float = 0.5):
        super(PDELU, self).__init__()
        if theta == 1.0:
            raise ValueError(
                "theta cannot be 1.0, as it would cause a division by zero."
            )
        self.theta = theta
        self._power_val = 1.0 / (1.0 - self.theta)
        self.alpha = torch.nn.UninitializedParameter()
        self._num_channels = None

    def _initialize_parameters(self, x: torch.Tensor):
        if x.ndim < 2:
            raise ValueError(
                f"Input tensor must have at least 2 dimensions (N, C), but got shape {x.shape}"
            )

        num_channels = x.shape[1]
        self._num_channels = num_channels
        param_shape = [1] * x.ndim
        param_shape[1] = num_channels
        init_tensor = torch.zeros(param_shape) + 0.1
        self.alpha = torch.nn.Parameter(init_tensor)

    def forward(self, x: torch.Tensor):
        if self.alpha is None:
            self._initialize_parameters(x)

        zero = torch.tensor(0.0, device=x.device, dtype=x.dtype)
        positive_part = torch.relu(x)
        inner_term = torch.relu(1.0 + (1.0 - self.theta) * x)
        powered_term = torch.pow(inner_term, self._power_val)
        subtracted_term = powered_term - 1.0
        negative_part = self.alpha * torch.min(subtracted_term, zero)
        return positive_part + negative_part



class HasNetwork:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # This property is going to be filled with the first model that is fed into the ensemble.
        self.network = None



from .CalculatesMetricNecessities import CalculatesMetricNecessities
from .HasNetwork import HasNetwork
from .PredictsProbabilities import PredictsProbabilities
from .RecordsStateDictionaries import RecordsStateDictionaries
from .ReportsClassificationMetrics import ReportsClassificationMetrics
from .ReportsEnsembleAccuracy import ReportsEnsembleAccuracy
from .ReportsEnsembleF1 import ReportsEnsembleF1
from .ReportsEnsembleKappa import ReportsEnsembleKappa
from .ReportsLogits import ReportsLogits
from .ReportsSize import ReportsSize



import pytest
import torch
from homa.vision import StochasticResnet, Model
from homa.vision.modules import ResnetModule


@pytest.fixture
def stochastic_resnet_model():
    return StochasticResnet(num_classes=5, lr=0.001)


def test_resnet_initialization(stochastic_resnet_model):
    assert isinstance(stochastic_resnet_model, StochasticResnet)
    assert isinstance(stochastic_resnet_model, Model)
    assert hasattr(stochastic_resnet_model, "network")
    assert hasattr(stochastic_resnet_model, "optimizer")
    assert hasattr(stochastic_resnet_model, "criterion")
    assert isinstance(stochastic_resnet_model.network, ResnetModule)
    assert isinstance(stochastic_resnet_model.optimizer, torch.optim.SGD)
    assert isinstance(stochastic_resnet_model.criterion, torch.nn.CrossEntropyLoss)



import torch


def replace_modules(
    model: torch.nn.Module, find: list | torch.Tensor, replacement: torch.nn.Module
) -> int:
    if not isinstance(find, list):
        find = [find]

    replaced = 0
    for parent in model.modules():
        for name, child in list(parent.named_children()):
            for needle in find:
                if isinstance(child, needle):
                    setattr(parent, name, replacement())
                    replaced += 1
    return replaced


def replace_relu(model: torch.nn.Module, replacement: torch.nn.Module):
    return replace_modules(model, torch.nn.ReLU, replacement)



import torch
import random

from .APLU import APLU
from .GALU import GALU
from .SmallGALU import SmallGALU
from .MELU import MELU
from .WideMELU import WideMELU
from .PDELU import PDELU
from .SReLU import SReLU


class StochasticActivation(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.gate = random.choice([APLU, GALU, SmallGALU, MELU, WideMELU, PDELU, SReLU])
        self.gate = self.gate()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.gate(x)



from copy import deepcopy
from .HasStateDicts import HasStateDicts
from ...vision import Model


class RecordsStateDictionaries(HasStateDicts):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def record(self, model: Model):
        if self.network is None:
            self.network = deepcopy(model.network)

        self.state_dicts.append(model.network.state_dict())

    def push(self, *args, **kwargs):
        self.record(*args, **kwargs)

    def append(self, *args, **kwargs):
        self.record(*args, **kwargs)

    def add(self, *args, **kwargs):
        self.record(*args, **kwargs)



from .ResnetModule import ResnetModule
from ..utils import replace_relu
from ...activations import StochasticActivation


class StochasticResnetModule(ResnetModule):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        replace_relu(self, StochasticActivation)



import torch


class SReLU(torch.nn.Module):
    def __init__(
        self,
        alpha_init: float = 0.0,
        beta_init: float = 0.0,
        gamma_init: float = 1.0,
        delta_init: float = 1.0,
    ):
        super().__init__()
        self.alpha_init_val = alpha_init
        self.beta_init_val = beta_init
        self.gamma_init_val = gamma_init
        self.delta_init_val = delta_init
        self.alpha = torch.nn.UninitializedParameter()
        self.beta = torch.nn.UninitializedParameter()
        self.gamma = torch.nn.UninitializedParameter()
        self.delta = torch.nn.UninitializedParameter()

    def _initialize_parameters(self, x: torch.Tensor):
        if isinstance(self.alpha, torch.nn.UninitializedParameter):
            if x.dim() < 2:
                raise ValueError(
                    f"Input tensor must have at least 2 dimensions (N, C), but got {x.dim()}"
                )

            num_channels = x.shape[1]
            param_shape = [1] * x.dim()
            param_shape[1] = num_channels
            self.alpha = torch.nn.Parameter(
                torch.full(param_shape, self.alpha_init_val)
            )
            self.beta = torch.nn.Parameter(torch.full(param_shape, self.beta_init_val))
            self.gamma = torch.nn.Parameter(
                torch.full(param_shape, self.gamma_init_val)
            )
            self.delta = torch.nn.Parameter(
                torch.full(param_shape, self.delta_init_val)
            )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        self._initialize_parameters(x)
        start = self.beta + self.alpha * (x - self.beta)
        finish = self.delta + self.gamma * (x - self.delta)
        out = torch.where(x < self.beta, start, x)
        out = torch.where(x > self.delta, finish, out)
        return out



from typing import List
from collections import OrderedDict


class HasStateDicts:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.state_dicts: List[OrderedDict] = []



from .modules import StochasticResnetModule
from .Resnet import Resnet


class StochasticResnet(Resnet):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.model = StochasticResnetModule(kwargs["num_classes"])



import torch
from .ReportsLogits import ReportsLogits


class PredictsProbabilities(ReportsLogits):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def predict(self, x: torch.Tensor) -> torch.Tensor:
        logits = self.logits(x)
        return torch.nn.functional.softmax(logits, dim=1)



import pytest
from homa.ensemble import Ensemble
from homa.vision import Resnet


@pytest.fixture
def ensemble():
    return Ensemble()


@pytest.fixture
def resnet():
    return Resnet(lr=0.001, num_classes=10)


def test_ensemble_initialization(ensemble):
    assert isinstance(ensemble, Ensemble)


def test_ensemble_records_models(ensemble, resnet):
    assert ensemble.network is None
    ensemble.record(resnet)
    assert ensemble.network is not None
    ensemble.append(resnet)
    ensemble.push(resnet)
    ensemble.add(resnet)
    assert ensemble.size == 4
    assert ensemble.length == 4



import torch


class CalculatesMetricNecessities:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def metric_necessities(self, dataloader):
        all_predictions = []
        all_labels = []
        for x, y in dataloader:
            batch_logits_list = []
            for model in self.models:
                batch_logits_list.append(model(x))
            all_batch_logits = torch.stack(batch_logits_list)
            avg_logits = torch.mean(all_batch_logits, dim=0)
            _, preds = torch.max(avg_logits, 1)
            all_predictions.extend(preds.cpu().numpy())
            all_labels.extend(y.cpu().numpy())
        return all_predictions, all_labels
